﻿using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace CrossWordCSharp
{
    class Algo
    {
        const double TIMELIMIT = 9.0;
        Crossword.MyServiceClient client;
        int w, h;
        int moves;
        char[,] field;
        RandomGen rnd;
        int stoneX, stoneY;
        IndexTree T;
        List<Tuple<int, int, int>> stonesSorted;
        int currentStoneIndex;
        System.DateTime clk0;
        int maxPlaceAttempts;
        int totalScore, badPlacements;

        public Algo(Crossword.MyServiceClient client)
        {
            this.client = client;
        }
        double btime()
        {
            System.TimeSpan t = System.DateTime.Now - clk0;
            return t.TotalMilliseconds / 1000.0;
        }
        static char prettyChar(char c)
        {
            if (c <= '1') return (c == '0') ? ' ' : '*';
            return c;
        }
        void determineStoneCorner()
        {
            // determine which quadrant has most stones initially, and use that for stone storage
            int[,] counts = new int[2, 2];
            for (int i = 0; i < h; i++)
                for (int j = 0; j < w; j++)
                    if (field[i, j] == '1')
                        counts[2 * i / h, 2 * j / w]++;
            int theMax = Math.Max(Math.Max(Math.Max(counts[0, 0], counts[0, 1]), counts[1, 0]), counts[1, 1]);
            for (int i = 0; i < 2; i++)
                for (int j = 0; j < 2; j++) 
                    if (counts[i, j] == theMax) {
                        stoneX = j * (w - 1);
                        stoneY = i * (h - 1);
#if DEBUG
                        Console.WriteLine("Stone corner set to (y = {0}, x = {1})", stoneY, stoneX);
#endif
                        return;
                    }
        }
        static int manhattan(int x1, int y1, int x2, int y2)
        {
            return Math.Abs(x1 - x2) + Math.Abs(y1 - y2);
        }
        void printField()
        {
            //return;
#if DEBUG
            for (int y = 0; y < h; y++) {
                string s = "    ";
                for (int x = 0; x < w; x++) s += prettyChar(field[y, x]);
                Console.WriteLine(s);
            }
#endif
        }
        bool makeStoneMove(int curMove)
        {
            if (currentStoneIndex >= stonesSorted.Count) return false;

            int x = stonesSorted[currentStoneIndex].Item2, y = stonesSorted[currentStoneIndex].Item3;
            currentStoneIndex += 1;
            var fc = T.findClosestFreeCell(stoneX, stoneY);
            int dx = fc.Item1, dy = fc.Item2;
            if (manhattan(stoneX, stoneY, dx, dy) >= manhattan(stoneX, stoneY, x, y)) return false;
            //
#if DEBUG
            Console.WriteLine("Move #{0}: Moving stone ({1}, {2}), current manhattan = {3} to ({4}, {5}) (score = {6})",
                curMove, y, x, manhattan(stoneX, stoneY, x, y), dy, dx, manhattan(stoneX, stoneY, dx, dy));
#endif
            Debug.Assert(field[dy, dx] == '0');
            Debug.Assert(field[y, x] == '1');
            field[dy, dx] = '1';
            field[y, x] = '0';
            T.update(dx, dy, +1);
            T.update(x, y, -1);
            client.relocate(y, x, dy, dx);
            return true;
        }
        public void run(int testcase)
        {
            rnd = new RandomGen();
            client.Init(testcase);
            totalScore = badPlacements = 0;

            w = client.getColumnCount();
            h = client.getRowCount();
            moves = client.getQueryCount();
            // Estimate how much test placement can we do per word.
            // 20 M ops for 10 seconds (being conservative);
            // 25 ops to check for fit
            // 
            maxPlaceAttempts = 20000000 / 25 / moves;
            maxPlaceAttempts = Math.Min(100000, maxPlaceAttempts);
            //
            field = new char[h, w];
            string[] ff = client.getField();
            for (int i = 0; i < h; i++)
                for (int j = 0; j < w; j++)
                    field[i, j] = ff[i][j];
            //
#if DEBUG
            Console.WriteLine("Width = {0}, Height = {1}, Moves = {2}", w, h, moves);
#endif

            T = new IndexTree(w, h);
            determineStoneCorner();
            stonesSorted = new List<Tuple<int, int, int>>();
            for (int y = 0; y < h; y++)
                for (int x = 0; x < w; x++)
                    if (field[y, x] == '1') {
                        stonesSorted.Add(new Tuple<int, int, int>(manhattan(x, y, stoneX, stoneY), x, y));
                        T.update(x, y, +1);
                    }
            stonesSorted.Sort();
            stonesSorted.Reverse();
            currentStoneIndex = 0;
            Queue<string> nextWords = new Queue<string>();
            clk0 = System.DateTime.Now;
            int halver = 0;
            //
            for (int i = 0; i < moves; i++) {
                if (moves % 20 == 0 && btime() > TIMELIMIT) break;
                if (nextWords.Count == 0 && (i == 0 || T.getFillRate() > 0.6)) {
                    string[] temp = client.skip();
                    foreach (string s in temp)
                        nextWords.Enqueue(s);
#if DEBUG
                    Console.WriteLine("Move #{0} was spent to get a list of words.", i);
#endif
                    continue;
                }
                if (nextWords.Count == 0) {
                    // don't have future info, cope with just hoping we get lucky:
                    string w1 = client.nextWord();
                    var r1 = planWord(w1);
#if DEBUG
                    Console.WriteLine("Move #{0}: blindly putting word {1} ({2}).", i, w1, (r1.Item1 > 0) ? "Success" : "Fail");
#endif
                    putWord(w1, r1.Item1 > 0, r1.Item2, r1.Item3, r1.Item4);
                    continue;
                }
                string word = nextWords.Peek();
                var res = planWord(word);
                if (nextWords.Count > 0 && res.Item1 == 0) {
                    // cannot make progress by placing the next word, try some stone-moving half of the time
                    if (halver++ % 2 == 0 && makeStoneMove(i)) {
                        printField();
                    } else {
                        // we chose not to (or just can't) move a stone. How sad. Just waste the turn fetching the next words:
                        nextWords.Clear();
                        string[] temp = client.skip();
                        foreach (string s in temp)
                            nextWords.Enqueue(s);
#if DEBUG
                        Console.WriteLine("Move #{0} was stalled, so fetched next words instead.", i);
#endif
                    }
                } else {
                    word = nextWords.Dequeue();
                    // have to place a word; either because it is beneficial, or because it's a bad
                    string redundant = client.nextWord();
                    Debug.Assert(redundant == word);
                    putWord(word, res.Item1 > 0, res.Item2, res.Item3, res.Item4);
#if DEBUG
                    Console.WriteLine("After move #{0}, word was \"{1}\", score = {2}:", i, word, res.Item1);
#endif
                    printField();
                }
            }
            //
            long time;
            client.EndGame(out time);
#if DEBUG
            Console.WriteLine("Elapsed time: {0} ms", time);
            Console.WriteLine("Total work time: {0} ms", (System.DateTime.Now - clk0).TotalMilliseconds);
            Console.WriteLine("Total score: {0}", totalScore);
#endif
            
        }
        Tuple<int, int, int, int> planWord(string word) // <score, sx, sy, vertical>
        {
            int l = word.Length;
            int bsx = -1, bsy = -1, bvertical = -1;
            int sx = -1, sy = -1, vertical = -1;
            int bestscore = -1;
            for (int tries = 0; tries < maxPlaceAttempts; tries++) {
                if (l > w) // has to be vertical
                    vertical = 1;
                else if (l > h) // has to be horizontal
                    vertical = 0;
                else
                    vertical = rnd.rand() % 2;
                if (vertical == 0) {
                    // horizontal
                    sx = rnd.rand() % (w - l + 1);
                    sy = rnd.rand() % h;
                } else {
                    sx = rnd.rand() % w;
                    sy = rnd.rand() % (h - l + 1);
                }
                int score = 0;
                for (int i = 0; i < l; i++) {
                    int x = sx + (vertical == 1 ? 0 : i);
                    int y = sy + (vertical == 1 ? i : 0);
                    if (field[y, x] == word[i]) score += 2;
                    else if (field[y, x] == '0') score++;
                    else {
                        score = 0; break;
                    }
                }
                if (score > bestscore) {
                    bestscore = score;
                    bsx = sx;
                    bsy = sy;
                    bvertical = vertical;
                }
            }

            if (bestscore > 0)
                return Tuple.Create(bestscore, bsx, bsy, bvertical);
            else {
                Debug.Assert(sx != -1);
                return Tuple.Create(0, sx, sy, vertical);
            }
        }
        void putWord(string word, bool doPlace, int sx, int sy, int vertical)
        {
            if (doPlace) {
                int l = word.Length;
                totalScore += l;
                for (int i = 0; i < l; i++) {
                    int x = sx + (vertical == 1 ? 0 : i);
                    int y = sy + (vertical == 1 ? i : 0);
                    if (field[y, x] != '0') totalScore++;
                    field[y, x] = word[i];
                    T.update(x, y, +1);
                }
            } else {
                if ((badPlacements++ % 2) == 1) totalScore -= 3;
            }
            bool result = client.placeWord(sy, sx, vertical == 1);
            Debug.Assert(doPlace == result);
        }
    }
}
